/*
hashmap narrative = [
    ("1", ("the sun shone brightly",      vector("A", "B"))
    ("2", ("you get punched in the eye!", vector("C", "D"))
    ("3", ("you stroke the cat!",         vector("1"))
]

hashmap decisions = [
    ("A", ("Kill him",       "1")
    ("B", ("Eat it",         "2")
    ("C", ("Yes",            "1")
    ("D", ("Stroke the cat", "3")
]
*/
// Please see stories/example.story for syntax examples

use std::fs;
use std::io;
use std::io::{stdin,stdout,Write};
use std::env;
use std::collections::HashMap;
use regex::Regex;

struct ParseError {
    line: i64,
    error_type: String,
}

fn clear_screen() {
    println!("Clearing Screen");
    clearscreen::clear().expect("Failed to clear screen"); 
}

fn parse_story(file_contents: &String) -> 
Result<
    HashMap<String, (i64, String, Vec<String>)>, 
    ParseError
>
{
    // Hashmaps for storing the contents of the story
    let mut narrative: HashMap<String, (i64, String, Vec<String>)> = HashMap::new();

    // Define the regex patterns of a peice of narrative and a decision look like
    let narrative_regex = Regex::new(r"[\s|\t|\r]?(\d+|[A-Za-z]+) (.*) \{(.*)\}").unwrap(); // eg: "17 You to Greggs {A, B}"
    // Parse the file contents in to the two hashmaps
    for (line_number, line_string) in file_contents.lines().enumerate()
    {
        // Parse a narrative line
        if narrative_regex.is_match(line_string)
        {
            for cap in narrative_regex.captures_iter(line_string)
            {
                let list_string = cap[3].replace(&" ", "");
                // Split the list into a vector
                let list: Vec<String> = list_string.split(",").map(str::to_string).collect();
                narrative.insert(cap[1].to_string(), (line_number as i64 + 1, cap[2].to_string(), list));
            }
        }
    }
    // Check that the file is valid
    if ! narrative.contains_key("0")
    {
        return Err(ParseError
            {
                line: 0,
                error_type: "No Narrative with index of 0, Therefore the narrative cannot begin".to_string(),
            })
    }
    return Ok(narrative)
}

fn input() -> String{
    let mut s=String::new();
    let _=stdout().flush();
    stdin().read_line(&mut s).expect("Did not enter a correct string");
    if let Some('\n')=s.chars().next_back() {
        s.pop();
    }
    if let Some('\r')=s.chars().next_back() {
        s.pop();
    }
    return s;
}

fn main() -> io::Result<()> 
{
    // Parse the arguments
    let args: Vec<String> = env::args().collect();
    
    if args.len() <= 1
    {
        println!("ERROR: Please provide the file path of the interactive story\n$ interactive-story -- story.txt");
    }
    else if &args[1] == "help"
    {
        println!("Help text here");
    }
    else
    {
        // Read the contents of the story to string
        let story_file_contents: String = fs::read_to_string(&args[1])?;

        // Parse the story
        let result = parse_story(&story_file_contents);
        match result
        {
            Err(error) => println!("Erorr on line {}:\n{}", error.line, error.error_type),
            Ok(narrative) => {
                clear_screen();
                let mut indexs: Vec<String> = vec!{"0".to_string()};
                let mut line: i64 = 0;
                'gameloop: loop
                {
                    // We should check if any of the index is "exit", if so
                    // that means we should quit, but only if there's exactly one index
                    if indexs[0].to_lowercase() == "exit" && indexs.len() == 1
                    {
                        break 'gameloop;
                    }
                    // We must check that all the indexs actually exist
                    let mut all_the_indexs_exist = true;
                    let mut none_existant_indexs = Vec::new();
                    for index in &indexs
                    {
                        // Check if narrative contains the key that is the current index
                        if ! narrative.contains_key(index)
                        {
                            all_the_indexs_exist = false;
                            none_existant_indexs.push(index);
                        }
                    } // for index in &indexs {

                    if all_the_indexs_exist 
                    {
                        // If there's only one index then it's just a peice of narration, so we ouput
                        // it and store that one's indexs
                        if indexs.len() == 1
                        {
                            let current_narrative = narrative.get(&indexs[0]).unwrap();
                            indexs = current_narrative.2.clone();
                            line = current_narrative.0;
                            println!("{}", current_narrative.1);
                        }
                        // However, if there are more than one index, then that means it is a decision
                        // step and so we prompt the player to make a decision
                        else if indexs.len() > 1
                        {
                            // relative_index is just the relative number, always starts at 0, it's used as
                            // a reference for the user to input.
                            // narrative_index is the actual index of the option that can be chosen
                            for (relative_index, narrative_index) in indexs.iter().enumerate()
                            {
                                let decision_text = &narrative.get(narrative_index).unwrap().1;
                                println!("{}: {}", relative_index, decision_text);
                            }
                            let choice: String = input();
                            let choice_index: usize = choice.trim().parse().unwrap(); // TODO potential crash?
                            line = narrative.get(&indexs[choice_index]).unwrap().0;
                            indexs = narrative.get(&indexs[choice_index]).unwrap().2.clone();
                        }
                    } // if all_the_indexs_exist {
                    else
                    {
                        println!("Error on line {}, narrative(s) of index(s) {:?} does not exist", line, none_existant_indexs);
                        break 'gameloop;
                    } // else {
                    println!(""); // Just print a newline to make it look nicer
                } // 'gameloop: loop {
            } // Ok(narrative, decision) => {
        } // match result {
    } // else {
    
    println!("Exiting...");
    Ok(())
}
